#!/usr/bin/env python3

import asyncio
import os
import sys

from fritzbox import FritzBox
from homeassistant import HomeAssistantAPI
import logging
import json

sensor_mappings = {}
thermostate_mappings = {}


async def handle_event(idx: int):
    logging.info(f"Wait for events for {idx}")

    while event := await ha.events[idx].get():
        try:
            entity_id = event["data"]["entity_id"]
            if entity_id in sensor_mappings.keys() or entity_id in thermostate_mappings.keys():
                state = await ha.get_device_state(entity_id)
                new_state = event["data"]["new_state"]
                logging.debug(f"received changed state from {entity_id}")
                if entity_id in thermostate_mappings.keys() and state["state"] != "unavailable":
                    therm_temp = new_state["attributes"]["current_temperature"]
                    therm_name = new_state["attributes"]["friendly_name"]
                    sensor = thermostate_mappings[entity_id]
                    sensor_state = await ha.get_device_state(sensor)
                    sensor_state_temp = sensor_state["attributes"]["temperature"] if "temperatur" in sensor_state[
                        "attributes"] else sensor_state["state"]
                    sensor_temp = round(float(sensor_state_temp) * 2) / 2
                    if therm_temp != sensor_temp:
                        logging.info(
                            f"{therm_name}: {therm_temp}\n{sensor}: {sensor_state_temp} ({sensor_temp})")
                        fb.correct_offset(therm_name, sensor_temp)

                elif entity_id in sensor_mappings.keys():
                    new_state_temp = new_state["attributes"]["temperature"] if "temperatur" in new_state[
                        "attributes"] else new_state["state"]
                    sensor_temp = round(float(new_state_temp) * 2) / 2
                    logging.info(sensor_temp)
                    for thermostate in sensor_mappings[entity_id]:
                        therm_state = await ha.get_device_state(thermostate)
                        if therm_state["state"] == "unavailable":
                            continue
                        therm_temp = float(therm_state["attributes"]["current_temperature"])
                        therm_name = therm_state["attributes"]["friendly_name"]
                        if therm_temp != sensor_temp or True:
                            logging.info(
                                f"{therm_name}: {therm_temp}\n{entity_id}: {new_state_temp} ({sensor_temp})")
                            fb.correct_offset(therm_name, sensor_temp)
        except KeyError as e:
            logging.error(e)


async def init(ha: HomeAssistantAPI, fb: FritzBox):
    if not await ha.connect():
        return
    logging.debug("Subscribe")
    state_changed_id = await ha.subscribe_event("state_changed")
    logging.debug(state_changed_id)
    asyncio.create_task(handle_event(state_changed_id))
    fb.login()
    await ha.wait_for_close()
    logging.info("Websocket closed, shutting down..")


async def main():
    config_path = sys.argv[1]
    config = json.load(open(config_path))
    level = logging.INFO
    if "log_level" in config:
        if config["log_level"] == "DEBUG":
            level = logging.DEBUG
    logging.basicConfig(level=level, format="[%(asctime)s]  [%(levelname)s]  %(message)s")
    logging.debug(config)

    global fb
    fb = FritzBox(url=config["fritzbox"]["url"],
                  user=config["fritzbox"]["username"],
                  password=config["fritzbox"]["password"],
                  update_timeout=config["update_timeout"],
                  dry_run=False,
                  verify_ssl=config["fritzbox"]["verify_ssl"],
                  old_fb=config["fritzbox"]["old_fb"])
    supervisor_url = "ws://supervisor/core/websocket"
    supervisor_token = os.environ["SUPERVISOR_TOKEN"]
    global ha
    ha = HomeAssistantAPI(supervisor_token, supervisor_url)

    for mapping in config["mappings"]:
        if mapping["sensor"] not in sensor_mappings.keys():
            sensor_mappings[mapping["sensor"]] = []
        sensor_mappings[mapping["sensor"]].append(mapping["thermostate"])
        thermostate_mappings[mapping["thermostate"]] = mapping["sensor"]

    try:
        await init(ha, fb)
    except KeyboardInterrupt:
        pass


asyncio.run(main())
